import numpy as np
import tal
from tal.config import get_resources, get_memory_usage
from tqdm import tqdm


def backproject_pf_single_frequency_naive(H_0, d, t, frequency):
    """
    H_0: (nt, ns)
    d: (ns, nv)
    t: (nt)
    frequency: scalar, in meters
    """
    assert H_0.ndim == 2, 'Incorrect H format'

    ns, nv = d.shape
    nt, ns_ = H_0.shape
    assert ns == ns_, 'Incorrect shape'

    propagator = np.exp(2j * np.pi * d * frequency)

    e = np.exp(-2j * np.pi * t * frequency)
    H_0_w = np.sum(H_0 * e.reshape((nt, 1)), axis=0).reshape((ns, 1))
    H_1_w = np.sum(H_0_w * propagator, axis=0)

    return H_1_w


def backproject_pf_single_frequency_conv(H_0, d_014, d_2, d_3, t, frequency):
    """
    Special case where the volume is a slice parallel to the relay wall
    with the same resolution and XY positions (only Z varies)

    The convolution can be implemented in frequency domain and can be much faster

    H_0: (nt, nsx, nsy)
    d_014: scalar, in meters
    d_2: (nsx, nsy)
    d_3: (rx, ry) where rx = 2 * nsx - 1 and ry = 2 * nsy - 1
    t: (nt)
    frequency: scalar, in meters
    """
    assert H_0.ndim == 3, 'Incorrect H format'

    rx, ry = d_3.shape
    nt, nsx, nsy = H_0.shape
    assert rx == 2 * nsx - 1 and ry == 2 * nsy - 1, 'Incorrect shape'

    rsd_014 = np.exp(2j * np.pi * d_014 * frequency)
    rsd_3 = np.exp(2j * np.pi * d_3 * frequency)

    e = np.exp(-2j * np.pi * t * frequency)
    H_0_w = np.sum(H_0 * e.reshape((nt, 1, 1)), axis=0).reshape((nsx, nsy))

    H_0_w *= rsd_014

    H_0_w_fft = np.fft.fft2(H_0_w, s=(rx, ry))
    rsd_3_fft = np.fft.fft2(np.fft.ifftshift(rsd_3), s=(rx, ry))
    H_1_w_full_fft = H_0_w_fft * rsd_3_fft
    H_1_w_full = np.fft.ifft2(H_1_w_full_fft)
    H_1_w = H_1_w_full[:nsx, :nsy]

    if d_2 is not None:
        rsd_2 = np.exp(2j * np.pi * d_2 * frequency)
        H_1_w *= rsd_2

    return H_1_w


def backproject_pf_multi_frequency(
        H_0, laser_grid_xyz, sensor_grid_xyz, volume_xyz,
        camera_system, t_accounts_first_and_last_bounces,
        t_start, delta_t,
        wl_mean, wl_sigma, border,
        laser_xyz=None, sensor_xyz=None, progress=False):
    assert laser_grid_xyz.size == 3, 'Only supports one laser position'
    assert (not t_accounts_first_and_last_bounces or (laser_xyz is not None and sensor_xyz is not None)), \
        't_accounts_first_and_last_bounces requires laser_xyz and sensor_xyz'

    nt = H_0.shape[0]
    nv = np.prod(volume_xyz.shape[:-1])  # N or X * Y or X * Y * Z
    ns = np.prod(sensor_grid_xyz.shape[:-1])  # N or X * Y

    try:
        assert H_0.ndim == 3
        assert sensor_grid_xyz.ndim == 3
        assert volume_xyz.ndim == 3 or volume_xyz.ndim == 4
        assert not t_accounts_first_and_last_bounces

        nt, nsx, nsy = H_0.shape
        nsx_, nsy_ = sensor_grid_xyz.shape[:2]
        assert nsx == nsx_ and nsy == nsy_
        nvx, nvy = volume_xyz.shape[:2]
        assert nvx == nsx and nvy == nsy

        def check_parallel(xyz_a, xyz_b):
            return np.allclose(xyz_a[..., 0:2], xyz_b[..., 0:2])

        if volume_xyz.ndim == 3:
            assert check_parallel(sensor_grid_xyz, volume_xyz)
        elif volume_xyz.ndim == 4:
            nvx, nvy, nvz = volume_xyz.shape[:3]
            for iz in range(nvz):
                depths = volume_xyz[:, :, iz, 2].reshape(-1)
                assert np.all(depths == depths[0])
                assert check_parallel(sensor_grid_xyz, volume_xyz[:, :, iz])

        print('tal.reconstruct.pf_dev: Using convolutions optimization')
        optimize_slices = True
    except AssertionError:
        assert np.prod(H_0.shape[1:]) == ns, 'Incorrect shape'
        H_0 = H_0.reshape((nt, ns))
        sensor_grid_xyz = sensor_grid_xyz.reshape((ns, 3))
        volume_xyz = volume_xyz.reshape((nv, 3))
        optimize_slices = False

    """ Phasor fields filter """

    # FIXME(diego): obtain filter parameters directly in frequency
    # domain (avoid conversion from time domain)

    t_6sigma = int(np.ceil(6 * wl_sigma / delta_t))
    padding = 2 * t_6sigma
    # FIXME(diego) if we want to convert a circular convolution to linear,
    # this should be nt + t_6sigma - 1 instead of nt + 4 * t_6sigma or even nt + 2 * t_6sigma
    # I have found cases where it fails even with nt + 2 * t_6sigma (Z, 0th pixel)
    nf = nt + 2 * padding

    t_max = delta_t * (nf - 1)
    t = np.linspace(start=0, stop=t_max, num=nf)

    gaussian_envelope = np.exp(-((t - t_max / 2) / wl_sigma) ** 2 / 2)
    K = gaussian_envelope / np.sum(gaussian_envelope) * \
        np.exp(2j * np.pi * t / wl_mean)
    K = np.fft.fftshift(K)  # center at zero

    mean_idx = (nf * delta_t) / wl_mean
    sigma_idx = (nf * delta_t) / (wl_sigma * 6)
    # shift to center at zero, easier for low negative frequencies
    freq_min_idx = nf // 2 + int(np.floor(mean_idx - 3 * sigma_idx))
    freq_max_idx = nf // 2 + int(np.ceil(mean_idx + 3 * sigma_idx))
    K_fftshift = np.fft.fftshift(np.fft.fft(K))
    K_fftfreq = np.fft.fftshift(np.fft.fftfreq(nf, d=delta_t))

    weights = K_fftshift[freq_min_idx:freq_max_idx+1]
    freqs = K_fftfreq[freq_min_idx:freq_max_idx+1]
    print('tal.reconstruct.pf_dev: '
          f'Using wavelengths from {1 / freqs[-1]:.4f}m to {1 / freqs[0]:.4f}m')
    nw = len(weights)

    if border == 'zero':
        # only pad temporal dimension
        H_0 = np.pad(H_0,
                     ((padding, padding),) + ((0, 0),) * (H_0.ndim - 1), 'constant')
    else:
        raise AssertionError('Implemented only for border="zero"')

    """ Propagation of specific frequencies """

    if camera_system.is_transient():
        H_1 = np.zeros((nt, nv), dtype=np.complex64)
    else:
        H_1 = np.zeros(nv, dtype=np.complex64)

    d_014 = t_start * -1
    if t_accounts_first_and_last_bounces:
        # d_1
        d_014 += np.linalg.norm(laser_xyz.reshape(3) -
                                laser_grid_xyz.reshape(3))
        # d_4
        d_014 += np.linalg.norm(sensor_xyz.reshape((1, 1, 3)) -
                                sensor_grid_xyz.reshape((ns, 1, 3)), axis=2)

    t = delta_t * np.linspace(start=0, stop=nf - 1, num=nf)

    h0 = H_0.dtype.itemsize
    h1 = H_1.dtype.itemsize
    s = sensor_grid_xyz.dtype.itemsize

    if optimize_slices:
        # optimized for when the volume is formed
        # with XY slices parallel to the relay wall
        if volume_xyz.ndim == 3:
            nvx, nvy, _ = volume_xyz.shape
            volume_xyz = volume_xyz.reshape((nvx, nvy, 1, 3))
        nvx, nvy, nvz, _ = volume_xyz.shape

        old_H_1_shape = H_1.shape
        if camera_system.is_transient():
            H_1 = H_1.reshape((nt, nvx, nvy, nvz))
        else:
            H_1 = H_1.reshape((nvx, nvy, nvz))

        range_z = np.arange(nvz, dtype=np.int32)

        dx = volume_xyz[1, 0, 0, 0] - volume_xyz[0, 0, 0, 0]
        dy = volume_xyz[0, 1, 0, 1] - volume_xyz[0, 0, 0, 1]
        rx = 2 * nvx - 1
        ry = 2 * nvy - 1
        px = np.linspace(start=-dx * (nvx - 1), stop=dx *
                         (nvx - 1), num=rx, dtype=np.float32)
        px = np.squeeze(np.stack((px,)*ry, axis=1))
        py = np.linspace(start=-dy * (nvy - 1), stop=dy *
                         (nvy - 1), num=ry, dtype=np.float32)
        py = np.squeeze(np.stack((py,)*rx, axis=0))

        def work_conv(subrange_z):
            volume_xyz_d = volume_xyz[..., subrange_z, :]
            nvx, nvy, nvzd, _ = volume_xyz_d.shape

            if camera_system.is_transient():
                H_1 = np.zeros((nt, nvx, nvy, nvzd), dtype=np.complex64)
            else:
                H_1 = np.zeros((nvx, nvy, nvzd), dtype=np.complex64)

            z_list = volume_xyz_d[0, 0, :, 2]
            if progress:
                progress_bar = tqdm(leave=False,
                                    total=len(z_list) * len(freqs))

            for iz, pz in enumerate(z_list):
                progress_bar.set_description(f'Z = {pz:.2f}m')
                d_2 = None
                if camera_system.bp_accounts_for_d_2():
                    d_2 = np.linalg.norm(
                        volume_xyz_d[:, :, iz, :] -
                        laser_grid_xyz.reshape((1, 1, 3)),
                        axis=2)
                d_3 = np.linalg.norm(
                    np.stack(
                        (px, py, np.ones((rx, ry), dtype=np.float32) * pz),
                        axis=2),
                    axis=2)

                fw_iterator = zip(freqs, weights)

                for frequency, weight in fw_iterator:
                    H_1_w = weight * \
                        backproject_pf_single_frequency_conv(
                            H_0, d_014, d_2, d_3, t, frequency)
                    e = np.exp(2j * np.pi * t * frequency) / nf
                    H_1_i = H_1_w.reshape((1, nvx, nvy)) * \
                        e.reshape((nf, 1, 1))

                    if camera_system.is_transient():
                        H_1[..., iz] += H_1_i[padding:-padding, ...]
                    else:
                        H_1[..., iz] += H_1_i[padding, ...]

                    if progress:
                        progress_bar.update(1)

            return H_1

        get_resources().split_work(
            work_conv,
            data_in=range_z,
            data_out=H_1,
            f_mem_usage=lambda dc: (
                lambda downscale, cpus:
                get_memory_usage(
                    ((ns, 3), s * (1 + cpus)), ((4 * ns, 3), s * (1 + cpus)),
                    (H_0.shape, (1 + h0) * cpus), (H_1.shape, 2 * h1))
            )(*dc),
            slice_dims=(0, H_1.ndim - 1),
        )

        H_1 = H_1.reshape(old_H_1_shape)
    else:
        # arbitrary volume
        range_v = np.arange(nv, dtype=np.int32)

        def work_n3(subrange_v):
            nvd = len(subrange_v)
            volume_xyz_d = volume_xyz[subrange_v]

            if camera_system.is_transient():
                H_1 = np.zeros((nt, nvd), dtype=np.complex64)
            else:
                H_1 = np.zeros(nvd, dtype=np.complex64)

            # d_3
            d = np.linalg.norm(
                sensor_grid_xyz.reshape((ns, 1, 3)) -
                volume_xyz_d.reshape((1, nvd, 3)),
                axis=2)
            if camera_system.bp_accounts_for_d_2():
                # d_2
                d += np.linalg.norm(
                    laser_grid_xyz.reshape((1, 1, 3)) -
                    volume_xyz_d.reshape((1, nvd, 3)),
                    axis=2)
            d += d_014  # add t_start, d_1 and d_4 terms

            iterator = zip(freqs, weights)
            if progress:
                iterator = tqdm(iterator, total=nw, leave=False)

            for frequency, weight in iterator:
                H_1_w = weight * \
                    backproject_pf_single_frequency_naive(
                        H_0, d, t, frequency)
                e = np.exp(2j * np.pi * t * frequency) / nf
                H_1_i = H_1_w.reshape((1, nvd)) * e.reshape((nf, 1))

                if camera_system.is_transient():
                    H_1 += H_1_i[padding: -padding, ...]
                else:
                    H_1 += H_1_i[padding]

            return H_1

        get_resources().split_work(
            work_n3,
            data_in=range_v,
            data_out=H_1,
            f_mem_usage=lambda dc: (
                lambda downscale, cpus:
                get_memory_usage(
                    ((ns, nv, 3), s * (1 + cpus / downscale)), (H_0.shape, (1 + h0) * cpus), (H_1.shape, 2 * h1))
            )(*dc),
            slice_dims=(0, H_1.ndim - 1),
        )

    return H_1



"""
tal.reconstruct.bp
==================

Reconstruction using the backprojection algorithm.
See "Recovering three-dimensional shape around a corner using ultrafast time-of-flight imaging."

Filtered backprojection is available in the fbp submodule, or by pre-filtering (see tal.reconstruct.filter_H).

Does _NOT_ attempt to compensate effects caused by attenuation:
    - cos decay i.e. {sensor|laser}_grid_normals are ignored
    - 1/d^2 decay
"""

from tal.io.capture_data import NLOSCaptureData
from tal.enums import VolumeFormat, CameraSystem
import numpy as np


def solve(data: NLOSCaptureData,
          volume_xyz: NLOSCaptureData.VolumeXYZType = None,
          volume_format: VolumeFormat = VolumeFormat.UNKNOWN,
          camera_system: CameraSystem = CameraSystem.DIRECT_LIGHT,
          progress: bool = True) -> np.array:  # FIXME(diego): volume type
    """
    See module description of tal.reconstruct.bp

    data
        See tal.io.read_capture

    volume_xyz
        Multi-dimensional array with XYZ coordinates of the volume voxels.
        See tal.enums.VolumeFormat for possible input formats.
        See e.g. tal.reconstruct.get_volume_min_max_resolution for utilities to generate this array.

    volume_format
        See tal.enums.VolumeFormat

    camera_system
        See tal.enums.CameraSystem

    progress
        If True, shows a progress bar with estimated time remaining.
    """
    from tal.reconstruct.utils import convert_to_N_3, convert_reconstruction_from_N_3
    H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3 = \
        convert_to_N_3(data, volume_xyz, volume_format)

    from tal.reconstruct.bp.backprojection import backproject
    reconstructed_volume_n3 = backproject(
        H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3,
        camera_system, data.t_accounts_first_and_last_bounces,
        data.t_start, data.delta_t,
        data.laser_xyz, data.sensor_xyz,
        progress=progress)

    return convert_reconstruction_from_N_3(data, reconstructed_volume_n3, volume_xyz, volume_format, camera_system)




from tal.io.capture_data import NLOSCaptureData
from typing import Union
import numpy as np
import pyvista as pv


def plot3d(data: Union[NLOSCaptureData, NLOSCaptureData.HType],
           title: str = '', slider_title: str = 'Time', 
           slider_step: float = 0.1, color = 'hot',
           opacity = 'sigmoid', backgroundcolor = None):
    """
    Plot in 3d the data indicated, with a slider if the data has a four 
    dimension
    @param data             : The data to plot. has to be 3 o 4 d
    @param title            : Title of the figure to plot. Default empty
    @param slider_tittle    : Title to print under the slider. Default "Time"
    @param color            : If string, it use that colormap to plot the 
                              image. It can be a 2d array of shape (N, 4), where
                              N are the number of points. Defalut "hot"
    @param opacity          : Opacity for the image. This argument is ignored
                              if color is an array.
    @param backgroundcolor  : Color of the background
    """
    data_to_plot = data
    if isinstance(data, NLOSCaptureData):
        data_to_plot = data.H
    
    assert data_to_plot.ndim >= 3 and data_to_plot.ndim <= 4,\
        "Expected 3 or 4 dimension data,"\
        +f" but data has dimension {data_to_plot.ndim}"
    
    p = pv.Plotter()
    p.add_title(title)
    if backgroundcolor is not None:
        p.background_color = backgroundcolor

    if isinstance(color, np.ndarray):
        cmap = None
        scalars = color
        if scalars.ndim > 2:
            scalars = scalars.reshape(-1, 4)
    else:
        scalars = None
        cmap = color

    # Normalize the data so it is possible to plot it with the plotter
    norm_data_to_plot = (data_to_plot - np.min(data_to_plot))* 256 \
                        / (np.max(data_to_plot) - np.min(data_to_plot))
    
    # Plot with slider if the data has more than 3 dimension
    if norm_data_to_plot.ndim == 4:
        assert cmap is not None, "Only colormaps can be used with a slider"
        def volume_by_time(slider_value):
            idx = int(slider_value/slider_step)
            p.add_volume(norm_data_to_plot[idx],
                        clim = [np.min(data_to_plot), np.max(data_to_plot)],
                        scalars = scalars,
                        cmap = cmap,
                        opacity = opacity,
                        name = 'data_by_time')

        p.add_slider_widget(volume_by_time, 
                            [0, (norm_data_to_plot.shape[0]-1) * slider_step],
                            value = 0,
                            title = slider_title)
    else: # 3 dim
        p.add_volume(norm_data_to_plot,
                    clim = [np.min(data_to_plot), np.max(data_to_plot)],
                    scalars = scalars,
                    cmap = cmap,
                    opacity = opacity,
                    name = 'full_data')

    # Select the initial camera
    p.camera_position = 'yz'
    p.camera.azimuth -= 15
    p.camera.elevation += 30
    # Plot the image
    p.show()



"""
tal.reconstruct.fbp
===================

Reconstruction using filtered backprojection.

Shortcut to call reconstruct.filter_H (pf filter) and reconstruct.bp.solve (see respective docstrings for details)
"""

from tal.io.capture_data import NLOSCaptureData
from tal.enums import VolumeFormat, CameraSystem
import numpy as np


def solve(data: NLOSCaptureData,
          wl_mean: float,
          wl_sigma: float,
          border: str = 'zero',
          volume_xyz: NLOSCaptureData.VolumeXYZType = None,
          volume_format: VolumeFormat = VolumeFormat.UNKNOWN,
          camera_system: CameraSystem = CameraSystem.DIRECT_LIGHT) -> np.array:  # FIXME(diego) type
    """
    See module description of tal.reconstruct.fbp

    data
        See tal.io.read_capture

    wl_mean, wl_sigma, border
        Filter parameters. See tal.reconstruct.filter_H

    volume_xyz
        Multi-dimensional array with XYZ coordinates of the volume voxels.
        See tal.enums.VolumeFormat for possible input formats.
        See e.g. tal.reconstruct.get_volume_min_max_resolution for utilities to generate this array.

    volume_format
        See tal.enums.VolumeFormat

    camera_system
        See tal.enums.CameraSystem

    progress
        If True, shows a progress bar with estimated time remaining.
    """
    from tal.reconstruct import filter_H
    old_H = data.H
    data.H = filter_H(data, filter_name='pf', border=border,
                      wl_mean=wl_mean, wl_sigma=wl_sigma)

    from tal.reconstruct.bp import solve as bp_solve
    H_1 = bp_solve(data,
                   volume_xyz=volume_xyz,
                   volume_format=volume_format,
                   camera_system=camera_system)
    data.H = old_H
    return H_1



from tal.enums import HFormat, GridFormat, VolumeFormat, CameraSystem
from tal.io.capture_data import NLOSCaptureData
import numpy as np


def _infer_volume_format(volume_xyz, volume_format, log=True):
    # infer volume_format if it is unknown
    if volume_format == VolumeFormat.UNKNOWN:
        assert volume_xyz.shape[-1] == 3, \
            'Could not infer volume_format. Please specify it manually.'
        if volume_xyz.ndim == 2:
            if log:
                print('tal.reconstruct.utils: Assuming that volume_xyz is N_3')
            volume_format = VolumeFormat.N_3
        elif volume_xyz.ndim == 3:
            if log:
                print('tal.reconstruct.utils: Assuming that volume_xyz is X_Y_3')
            volume_format = VolumeFormat.X_Y_3
        elif volume_xyz.ndim == 4:
            if log:
                print('tal.reconstruct.utils: Assuming that volume_xyz is X_Y_Z_3')
            volume_format = VolumeFormat.X_Y_Z_3
        else:
            raise AssertionError(
                'Could not infer volume_format. Please specify it manually.')

    return volume_format


def convert_to_N_3(data: NLOSCaptureData,
                   volume_xyz: NLOSCaptureData.VolumeXYZType,
                   volume_format: VolumeFormat = VolumeFormat.UNKNOWN,
                   try_optimize_convolutions: bool = False):

    volume_format = _infer_volume_format(volume_xyz, volume_format, log=True)

    # this variable is set to false during the conversion
    can_optimize_convolutions = \
        volume_format in [VolumeFormat.X_Y_3, VolumeFormat.X_Y_Z_3]

    if data.H_format == HFormat.T_Si:
        can_optimize_convolutions = False
        H = data.H
    elif data.H_format == HFormat.T_Sx_Sy:
        nt, nsx, nsy = data.H.shape
        H = data.H.reshape(nt, nsx * nsy)
    else:
        raise AssertionError(f'H_format {data.H_format} not implemented')

    # FIXME(diego) also convert laser data and confocal/exhaustive H measurements

    if data.sensor_grid_format == GridFormat.N_3:
        can_optimize_convolutions = False
        sensor_grid_xyz = data.sensor_grid_xyz
    elif data.sensor_grid_format == GridFormat.X_Y_3:
        try:
            assert nsx == data.sensor_grid_xyz.shape[0] and nsy == data.sensor_grid_xyz.shape[1], \
                'sensor_grid_xyz.shape does not match with H.shape'
        except NameError:
            # nsx, nsy not defined, OK
            nsx, nsy, _ = data.sensor_grid_xyz.shape
            pass
        sensor_grid_xyz = data.sensor_grid_xyz.reshape(nsx * nsy, 3)
    else:
        raise AssertionError(
            f'sensor_grid_format {data.sensor_grid_format} not implemented')

    assert H.shape[1] == sensor_grid_xyz.shape[0], \
        'H.shape does not match with sensor_grid_xyz.shape. Different number of points than measurements.'

    assert volume_format.xyz_dim_is_last(), 'Unexpected volume_format'
    try:
        assert try_optimize_convolutions and can_optimize_convolutions
        nvx, nvy = volume_xyz.shape[:2]
        assert nvx == nsx and nvy == nsy
        H = H.reshape((nt, nvx, nvy))
        sensor_grid_xyz = sensor_grid_xyz.reshape((nvx, nvy, 3))
        volume_xyz = volume_xyz
        print('tal.reconstruct.utils: Optimizing for convolutions. '
              'Other algorithms (e.g. tal.reconstruct.pf_dev) should also log that it is being used.')
    except AssertionError:
        volume_xyz = volume_xyz.reshape((-1, 3))

    return (H, data.laser_grid_xyz, sensor_grid_xyz, volume_xyz)


def convert_reconstruction_from_N_3(data: NLOSCaptureData,
                                    # FIXME(diego) type
                                    reconstructed_volume_n3: np.ndarray,
                                    volume_xyz: NLOSCaptureData.VolumeXYZType,
                                    volume_format: VolumeFormat,
                                    camera_system: CameraSystem):

    volume_format = _infer_volume_format(volume_xyz, volume_format, log=False)

    if camera_system.is_transient():
        time_dim = (data.H.shape[data.H_format.time_dim()],)
    else:
        time_dim = ()

    assert volume_format.xyz_dim_is_last(), 'Unexpected volume_format'
    return reconstructed_volume_n3.reshape(
        time_dim + volume_xyz.shape[:-1])



"""
tal.reconstruct.pf_dev
======================

Reconstruction using the phasor fields framework.
See "Non-Line-of-Sight Imaging using Phasor Field Virtual Wave Optics."

Does _NOT_ attempt to compensate effects caused by attenuation:
    - cos decay i.e. {sensor|laser}_grid_normals are ignored
    - 1/d^2 decay
"""

from tal.io.capture_data import NLOSCaptureData
from tal.enums import VolumeFormat, CameraSystem
import numpy as np


def solve(data: NLOSCaptureData,
          wl_mean: float,
          wl_sigma: float,
          border: str = 'zero',
          volume_xyz: NLOSCaptureData.VolumeXYZType = None,
          volume_format: VolumeFormat = VolumeFormat.UNKNOWN,
          camera_system: CameraSystem = CameraSystem.DIRECT_LIGHT,
          progress: bool = True,
          try_optimize_convolutions: bool = True) -> np.array:  # FIXME(diego) type
    """
    See module description of tal.reconstruct.pf_dev

    data
        See tal.io.read_capture

    wl_mean, wl_sigma, border
        Filter parameters. See tal.reconstruct.filter_H

    volume_xyz
        Multi-dimensional array with XYZ coordinates of the volume voxels.
        See tal.enums.VolumeFormat for possible input formats.
        See e.g. tal.reconstruct.get_volume_min_max_resolution for utilities to generate this array.

    volume_format
        See tal.enums.VolumeFormat

    camera_system
        See tal.enums.CameraSystem

    progress
        If True, shows a progress bar with estimated time remaining.

    try_optimize_convolutions
        When volume_xyz consists of depth-slices (Z-slices) that are parallel to the XY relay wall,
        the computation can be optimized to use less memory and be much faster.
        It is recommended to set this to True.
        You can generate these depth-slices with tal.reconstruct.get_volume_project_rw(...).
    """
    from tal.reconstruct.utils import convert_to_N_3, convert_reconstruction_from_N_3
    H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3 = \
        convert_to_N_3(data, volume_xyz, volume_format,
                       try_optimize_convolutions=try_optimize_convolutions)

    from tal.reconstruct.pf_dev.phasor_fields import backproject_pf_multi_frequency
    reconstructed_volume_n3 = backproject_pf_multi_frequency(
        H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3,
        camera_system, data.t_accounts_first_and_last_bounces,
        data.t_start, data.delta_t,
        wl_mean, wl_sigma, border,
        data.laser_xyz, data.sensor_xyz,
        progress=progress)

    return convert_reconstruction_from_N_3(data, reconstructed_volume_n3, volume_xyz, volume_format, camera_system)



import os
import shutil
import yaml
import tal
from tal.io.capture_data import NLOSCaptureData
from tal.enums import FileFormat, GridFormat, HFormat
from tal.config import local_file_path
import datetime
import numpy as np
from tqdm import tqdm

from tal.render.util import import_mitsuba_backend
mitsuba_backend = import_mitsuba_backend()


def render_nlos_scene(config_path, args):
    config_path = os.path.abspath(config_path)

    assert os.path.exists(config_path), \
        f'{config_path} does not exist'

    path, name = os.path.split(config_path)
    candidate_filename = os.path.join(path, name, f'{name}.yaml')
    if os.path.isdir(config_path) and os.path.isfile(candidate_filename):
        config_path = candidate_filename

    assert os.path.isfile(config_path), \
        f'{config_path} is not a TAL config file'

    config_dir, config_filename = os.path.split(config_path)

    try:
        scene_config = yaml.safe_load(
            open(config_path, 'r')) or dict()
        scene_defaults = yaml.safe_load(
            open(local_file_path('render/scene_defaults.yaml'), 'r'))
    except yaml.YAMLError as exc:
        raise AssertionError(
            f'Invalid YAML format in TAL config file: {exc}') from exc
    scene_config = {**scene_defaults, **scene_config}

    if not args.dry_run:
        mitsuba_backend.set_variant(scene_config['mitsuba_variant'])
    steady_xml, nlos_xml = mitsuba_backend.get_scene_xml(
        scene_config, random_seed=args.seed, quiet=args.quiet)

    try:
        root_dir = os.path.join(
            config_dir, datetime.datetime.now().strftime(r'%Y%m%d-%H%M%S'))
        partial_results_dir = os.path.join(root_dir, 'partial')
        steady_dir = os.path.join(root_dir, 'steady')
        log_dir = os.path.join(root_dir, 'logs')
        os.mkdir(root_dir)
        os.mkdir(partial_results_dir)
        os.mkdir(steady_dir)
        os.mkdir(log_dir)
        shutil.copy(
            config_path,
            os.path.join(root_dir, f'{config_filename}.old'))
    except OSError as exc:
        raise AssertionError(f'Invalid permissions: {exc}') from exc

    try:
        steady_scene_xml = os.path.join(root_dir, 'steady_scene.xml')
        with open(steady_scene_xml, 'w') as f:
            f.write(steady_xml)

        nlos_scene_xml = os.path.join(root_dir, 'nlos_scene.xml')
        with open(nlos_scene_xml, 'w') as f:
            f.write(nlos_xml)

        laser_lookats = []
        name = scene_config['name']
        scan_type = scene_config['scan_type']
        num_bins = scene_config['num_bins']
        sensor_width = scene_config['sensor_width']
        sensor_height = scene_config['sensor_height']
        laser_width = scene_config['laser_width']
        laser_height = scene_config['laser_height']
        if scan_type == 'single':
            laser_width = 1
            laser_height = 1

        relay_wall = next(filter(
            lambda g: g['name'] == scene_config['relay_wall'],
            scene_config['geometry']))
        assert 'displacement_x' not in relay_wall and \
            'displacement_y' not in relay_wall and \
            'displacement_z' not in relay_wall and \
            'rot_degrees_x' not in relay_wall and \
            'rot_degrees_y' not in relay_wall and \
            'rot_degrees_z' not in relay_wall, \
            'Relay wall displacement/rotation is NYI'

        def get_grid_xyz(nx, ny, relay_wall_scale):
            p = relay_wall_scale
            xg = np.stack((np.linspace(-p, p, num=2*nx + 1)[1::2],)*ny, axis=1)
            yg = np.stack((np.linspace(-p, p, num=2*ny + 1)[1::2],)*nx, axis=0)
            assert xg.shape[0] == yg.shape[0] == nx and xg.shape[1] == yg.shape[1] == ny, \
                'Incorrect shapes'
            return np.stack([xg, yg, np.zeros((nx, ny))], axis=-1).astype(np.float32)

        def expand(vec, x, y):
            assert len(vec) == 3
            return vec.reshape(1, 1, 3).repeat(x, axis=0).repeat(y, axis=1)

        # FIXME(diego): rotate + translate (asssumes no rot/trans)
        # or use a more generalist approach that does not need to be rectangular
        sensor_grid_xyz = get_grid_xyz(
            sensor_width, sensor_height, relay_wall['scale'])
        laser_grid_xyz = get_grid_xyz(
            laser_width, laser_height, relay_wall['scale'])
        # FIXME(diego): rotate [0, 0, 1] by rot_degrees_x (assmes RW is a plane)
        # or use a more generalist approach
        sensor_grid_normals = expand(
            np.array([0, 0, 1]), sensor_width, sensor_height)
        laser_grid_normals = expand(
            np.array([0, 0, 1]), laser_width, laser_height)

        if scan_type == 'single':
            laser_lookat_x = \
                scene_config['laser_lookat_x'] or (sensor_width - 1) / 2
            laser_lookat_y = \
                scene_config['laser_lookat_y'] or (sensor_height - 1) / 2
            laser_lookats.append((laser_lookat_x, laser_lookat_y))
        elif scan_type == 'exhaustive' or scan_type == 'confocal':
            assert not (scan_type == 'confocal' and
                        (laser_width != sensor_width or
                            laser_height != sensor_height)), \
                'If using scan_type=confocal, sensor_{width|height} must match laser_{width|height}'
            for x in range(laser_width):
                for y in range(laser_height):
                    laser_lookats.append((x + 0.5, y + 0.5))
        else:
            raise AssertionError(
                'Invalid scan_type, must be one of {single|exhaustive|confocal}')

        experiment_name = scene_config['name']

        if args.do_steady_renders:
            def render_steady(render_name, sensor_index):
                if not args.quiet:
                    print(f'{render_name} for {experiment_name} steady render...')
                hdr_ext = mitsuba_backend.get_hdr_extension()
                hdr_path = os.path.join(partial_results_dir,
                                        f'{experiment_name}_{render_name}.{hdr_ext}')
                ldr_path = os.path.join(steady_dir,
                                        f'{experiment_name}_{render_name}.png')
                logfile = None
                if args.do_logging and not args.dry_run:
                    logfile = open(os.path.join(
                        log_dir, f'{experiment_name}_{render_name}.log'), 'w')
                mitsuba_backend.run_mitsuba(steady_scene_xml, hdr_path, dict(),
                                            render_name, logfile, args, sensor_index)
                if args.do_logging and not args.dry_run:
                    logfile.close()
                if not args.dry_run:
                    mitsuba_backend.convert_hdr_to_ldr(hdr_path, ldr_path)

            render_steady('back_view', 0)
            render_steady('front_view', 1)

        for i, (laser_lookat_x, laser_lookat_y) in tqdm(
                enumerate(laser_lookats), desc=f'Rendering {experiment_name} ({scan_type})...',
                ascii=True, total=len(laser_lookats)):
            try:
                hdr_path, is_dir = mitsuba_backend.partial_laser_path(
                    partial_results_dir, experiment_name, laser_lookat_x, laser_lookat_y)
                if is_dir:
                    os.mkdir(hdr_path)
            except OSError as exc:
                raise AssertionError(f'Invalid permissions: {exc}') from exc
            defines = {
                'laser_lookat_x': laser_lookat_x,
                'laser_lookat_y': laser_lookat_y,
            }
            logfile = None
            if args.do_logging and not args.dry_run:
                logfile = open(os.path.join(
                    log_dir,
                    f'{experiment_name}_L[{laser_lookat_x}][{laser_lookat_y}].log'), 'w')
            mitsuba_backend.run_mitsuba(nlos_scene_xml, hdr_path, defines,
                                        f'Laser {i + 1} of {len(laser_lookats)}', logfile, args)
            if args.do_logging and not args.dry_run:
                logfile.close()

        if args.dry_run:
            return

        if not args.quiet:
            print('Merging partial results...')

        capture_data = NLOSCaptureData()
        capture_data.sensor_xyz = np.array([
            scene_config['sensor_x'],
            scene_config['sensor_y'],
            scene_config['sensor_z'],
        ], dtype=np.float32)
        capture_data.sensor_grid_xyz = sensor_grid_xyz
        capture_data.sensor_grid_normals = sensor_grid_normals
        capture_data.sensor_grid_format = GridFormat.X_Y_3
        capture_data.laser_xyz = np.array([
            scene_config['laser_x'],
            scene_config['laser_y'],
            scene_config['laser_z'],
        ], dtype=np.float32)
        capture_data.laser_grid_xyz = laser_grid_xyz
        capture_data.laser_grid_normals = laser_grid_normals
        capture_data.laser_grid_format = GridFormat.X_Y_3
        # NOTE(diego): we do not store volume information for now
        # capture_data.volume_format = VolumeFormat.X_Y_Z_3
        capture_data.delta_t = scene_config['bin_width_opl']
        capture_data.t_start = scene_config['start_opl']
        capture_data.t_accounts_first_and_last_bounces = \
            scene_config['account_first_and_last_bounces']
        capture_data.scene_info = {
            'tal_version': tal.__version__,
            'config': scene_config,
            'args': vars(args),
        }
        if scan_type == 'single':
            hdr_path, _ = mitsuba_backend.partial_laser_path(
                partial_results_dir,
                experiment_name,
                *laser_lookats[0])
            capture_data.H = mitsuba_backend.read_transient_image(hdr_path)
            capture_data.H_format = HFormat.T_Sx_Sy
        elif scan_type == 'exhaustive' or scan_type == 'confocal':
            if scan_type == 'exhaustive':
                capture_data.H = np.empty(
                    (num_bins, laser_width, laser_height,
                     sensor_width, sensor_height),
                    dtype=np.float32)
                capture_data.H_format = HFormat.T_Lx_Ly_Sx_Sy
            elif scan_type == 'confocal':
                capture_data.H = np.empty(
                    (num_bins, laser_width, laser_height),
                    dtype=np.float32)
                capture_data.H_format = HFormat.T_Sx_Sy
            else:
                raise AssertionError

            for x in range(laser_width):
                for y in range(laser_height):
                    hdr_path, _ = mitsuba_backend.partial_laser_path(
                        partial_results_dir,
                        experiment_name,
                        x + 0.5, y + 0.5)
                    capture_data.H[:, x, y, ...] = np.squeeze(
                        mitsuba_backend.read_transient_image(hdr_path))
        else:
            raise AssertionError(
                'Invalid scan_type, must be one of {single|exhaustive|confocal}')

        hdf5_path = os.path.join(root_dir, f'{experiment_name}.hdf5')
        tal.io.write_capture(capture_data, hdf5_path,
                             file_format=FileFormat.HDF5_TAL)

        if not args.quiet:
            print(f'Stored result in {hdf5_path}')

        if args.keep_partial_results:
            return

        if not args.quiet:
            print(f'Cleaning partial results in {partial_results_dir}...')

        shutil.rmtree(partial_results_dir)

        if not args.quiet:
            print(f'All clean.')
    except KeyboardInterrupt:
        delete = None
        while delete is None:
            try:
                answer = input(f'Render cancelled. '
                               f'Delete the directory {root_dir}? (y/n): ')
                if answer.lower() == 'y':
                    delete = True
                elif answer.lower() == 'n':
                    delete = False
            except KeyboardInterrupt:
                pass
        if delete:
            shutil.rmtree(root_dir)



from enum import Enum


class FileFormat(Enum):
    """
    File formats for NLOSCaptureData files.

    AUTODETECT
        Choose based on the file extension and contents.

    HDF5_ZNLOS
        HDF5 format used in Zaragoza-NLOS dataset
        https://graphics.unizar.es/nlos_dataset/

    HDF5_NLOS_DIRAC
        HDF5 format, deprecated

    HDF5_TAL
        HDF5 format generated by TAL (see tal render -h)

    MAT_PHASOR_FIELDS
        MAT format used in the work "Non-Line-of-Sight Imaging using Phasor Field Virtual Wave Optics"
        https://biostat.wisc.edu/~compoptics/phasornlos19/phasor_nlos_19.html
    """
    AUTODETECT = 0
    HDF5_ZNLOS = 1
    HDF5_NLOS_DIRAC = 2
    HDF5_TAL = 3
    MAT_PHASOR_FIELDS = 4


class HFormat(Enum):
    """
    Dimensions specification for impulse response data H.

    UNKNOWN
        Avoid using this value.

    T_Sx_Sy
        H is a 3D array with dimensions (T, Sx, Sy).
        This includes confocal and non-confocal data.

    T_Lx_Ly_Sx_Sy
        H is a 5D array with dimensions (T, Lx, Ly, Sx, Sy).

    T_Si
        H is a 2D array with dimensions (T, Si).
        This includes confocal and non-confocal data.

    T_Si_Li
        H is a 3D array with dimensions (T, Si, Li).
    """
    UNKNOWN = 0
    T_Sx_Sy = 1  # confocal or not
    T_Lx_Ly_Sx_Sy = 2
    T_Si = 3  # confocal or not
    T_Li_Si = 4

    def time_dim(self) -> int:
        """ Returns the index of the time (T) dimension. """
        assert self in [HFormat.T_Sx_Sy,
                        HFormat.T_Lx_Ly_Sx_Sy,
                        HFormat.T_Si,
                        HFormat.T_Li_Si], \
            f'Unexpected HFormat {self}'
        return 0


class GridFormat(Enum):
    """
    Dimensions specification for grid data.

    UNKNOWN
        Avoid using this value.

    N_3
        Grid is a 2D array with dimensions (N, 3).
        Prefer more specific formats (e.g. X_Y_3) if your data
        is in fact a 2D grid of points, there are some functions
        that optimize this fact.

    X_Y_3
        Grid is a 3D array with dimensions (X, Y, 3).
    """
    UNKNOWN = 0
    N_3 = 1
    X_Y_3 = 2


class VolumeFormat(Enum):
    """
    Dimensions specification for volume data.

    UNKNOWN
        Avoid using this value. Some implementations of the
        reconstruction algorithms will guess your volume format
        based on the data that you pass to it, but it is better
        to be explicit.

    N_3
        Volume is a 2D array with dimensions (N, 3).
        Prefer more specific formats (e.g. X_Y_3, X_Y_Z_3) if your data
        is in fact a 2D/3D set of points, there are some functions
        that optimize this fact.

    X_Y_Z_3
        Volume is a 4D array with dimensions (X, Y, Z, 3).

    X_Y_3
        Volume is a 3D array with dimensions (X, Y, 3).
    """
    UNKNOWN = 0
    N_3 = 1
    X_Y_Z_3 = 2
    X_Y_3 = 3

    def xyz_dim_is_last(self) -> bool:
        assert self in [VolumeFormat.N_3,
                        VolumeFormat.X_Y_Z_3,
                        VolumeFormat.X_Y_3]
        return True


class CameraSystem(Enum):
    """
    Reconstruction algorithms work as a virtual camera system.
    This camera system can focus at different points of the hidden scene,
    and virtually illuminate other points. The camera system chosen chooses
    the behaviour of the virtual camera.

    For more info., read the supplementary material of the paper
    "Non-Line-of-Sight Imaging using Phasor Field Virtual Wave Optics"
    https://biostat.wisc.edu/~compoptics/phasornlos19/phasor_nlos_19.html
    (section A and Tables S.2 and S.3)

    DIRECT_LIGHT
        Confocal camera at t = 0. Most common NLOS reconstruction system.

    CONFOCAL_TIME_GATED
        Confocal camera with time gating. Computes a video of the scene (t >= 0).
        The returned reconstruction will have an extra time dimension.

    TRANSIENT
        Transient camera. Computes a video of the scene (t >= 0).
        The returned reconstruction will have an extra time dimension.
    """
    DIRECT_LIGHT = 0  # focused light (evaluate confocal time gated at t=0)
    CONFOCAL_TIME_GATED = 1  # pulsed focused light
    # TRANSIENT_T0 = 2  # NYI (evaluate transient at t=0), also add to functions below
    # STEADY_STATE = 3  # NYI (integrate transient over time), also add to functions below
    TRANSIENT = 4  # pulsed point light
    # PHOTO_CAMERA = 5  # NYI (single-freq imaging), also add to functions below

    def bp_accounts_for_d_2(self) -> bool:
        return self in [CameraSystem.DIRECT_LIGHT, CameraSystem.CONFOCAL_TIME_GATED]

    def is_transient(self) -> bool:
        # NOTE: prob. need to divide into "returns_transient" and "computes_transient"
        # or sth like that
        return self in [CameraSystem.TRANSIENT, CameraSystem.CONFOCAL_TIME_GATED]



"""
Author      :   Pablo Luesia Lahoz
E-Mail      :   pluesia at unizar.es
File        :   propagator.py
Description :   Interface of propagation to use with Phasor Fields solver, and
                RSD implementations for Parallel and non Parallel planes
"""

import numpy as np
from threading import Lock
from tal.reconstruct.pf.rsd_kernel import RSD_kernel
from typing import Tuple, Union


class Propagator(object):
    """
    Propagation interface
    """

    def propagate(self, fH: np.ndarray, P: np.ndarray,
                  V: np.ndarray,  wl_v: np.ndarray,
                  P_axis: Union[int, Tuple[int]] = None,
                  V_axis: Union[int, Tuple[int]] = None) -> np.ndarray:
        """
        Propagate from P to V the value fH, where fH is the impulse response at
        P in the Fourier domain, and wl_v is the array of wavelengths for each
        fH component
        @param self     : Propagator subclass implementation
        @param fH       : Impulse response in the Fourier domain
        @param P        : Coordinates of the fH samples
        @param V        : Coordinates to propagate fH
        @param wl_v     : Array of wavelengths for each fH Fourier component
        @param P_axis   : Axis of fH which represents P. If None, axis is 
                          assumed to be the last one
        @param V_axis   : Axis of fH which represents V. If None, it is assumed
                          there is no axis representing V. It use the axis to
                          partially propagations (e.g. propagation from sensors
                          but not from light sources)
        @return         : The propagation of fH to the V coordinates
        """
        raise NotImplementedError('Subclasses should implement propagate')

    @staticmethod
    def reshape_data(fH: np.ndarray, P: np.ndarray, V: np.ndarray,
                     axis: Union[int, Tuple[int]]):
        """
        Reshape the data to the needed form for the propagation implementations
        """
        raise NotImplementedError('Subclasses should implement reshape_data')

    # Reorder the results to match the given data

    @staticmethod
    def reshape_result(fI: np.ndarray, fH: np.ndarray, V: np.ndarray,
                       w_axis: int,
                       P_axis: Union[int, Tuple[int]],
                       V_axis: Union[int, Tuple[int]]):

        # Reshape to the corresponding shape. By axis: frequency, fH
        # uncorresponding to P shape, V shape
        V_l_axis = V_axis
        if V_l_axis is None:
            I_shape = tuple(np.delete(np.array(fH.shape), P_axis)) \
                + tuple(np.array(V.shape)[:-1])
            V_l_axis = tuple(np.arange(- V.ndim + 1, 0))
        else:
            I_shape = tuple(np.delete(np.array(fH.shape), P_axis + V_axis)) \
                + tuple(np.array(V.shape)[:-1])
        # Move the axis of frequencies to the beginning
        fI_aux = np.moveaxis(fI, w_axis, 0)
        # Reshape the data
        I = fI_aux.reshape(I_shape)
        return I

    @staticmethod
    def check_shapes(fH: np.ndarray, P: np.ndarray,
                     V: np.ndarray,  wl_v: np.ndarray,
                     P_axis: Union[int, Tuple[int]],
                     V_axis: Union[int, Tuple[int]]):
        """
        Check the shapes of all the input data
        """
        # Reshape to arrays all the data
        V_shape = np.array(V.shape)
        P_shape = np.array(P.shape)
        fH_shape = np.array(fH.shape)

        assert np.all(fH_shape[np.array(P_axis)] == P_shape[:-1]), \
            "fH and P are not the same shape"
        if V_axis is not None:
            assert np.all(fH_shape[np.array(V_axis)] == V_shape[:-1]), \
                "Indicated axis for V in H, but shapes are not the same"
        assert wl_v.shape[0] == fH.shape[0], \
            'number of frequencies does not match in the first axis of fH'
        assert V_shape[-1] == 3, "propagate do not support this data format"
        assert P_shape[-1] == 3, "propagate do not support this data format"

        return (V_shape, P_shape, fH_shape)

    @staticmethod
    def axis_value(axis):
        """
        Return a value given axis. Iff None returns axis -1
        """
        if axis is None:
            return (-1,)
        else:
            return axis


class RSD_propagator(Propagator):
    """
    Propagation using Rayleigh-Sommerfeld Diffraction
    """

    def __init__(self):
        super().__init__()

    def propagate(self, fH: np.ndarray, P: np.ndarray,
                  V: np.ndarray,  wl_v: np.ndarray,
                  P_axis: Union[int, Tuple[int]] = None,
                  V_axis: Union[int, Tuple[int]] = None) -> np.ndarray:
        """
        Overrides propagate. It uses a RSD propagation. It collapses the 
        given axis, and add new ones of size V
        """
        # Extract the axis and check the shapes
        P_l_axis = RSD_propagator.axis_value(P_axis)
        RSD_propagator.check_shapes(fH, P, V, wl_v, P_l_axis, V_axis)
        # Reshape the data to the needed form
        V_a, P_a, fH_a = RSD_propagator.reshape_data(
            fH, P, V, P_l_axis, V_axis)
        # calculate distances from P to all V
        dist = np.linalg.norm(P_a - V_a, axis=-1)
        # RSD kernel calculation
        RSD_k = RSD_kernel.RSD_kernel_w(dist, wl_v)
        # Propagate the light, and focus on the voxel
        I_a = np.sum(fH_a*RSD_k, axis=-2)
        I = RSD_propagator.reshape_result(I_a, fH, V, -2, P_l_axis, V_axis)

        return I

    @staticmethod
    def reshape_data(fH: np.ndarray, P: np.ndarray, V: np.ndarray,
                     P_axis: Union[int, Tuple[int]],
                     V_axis: Union[int, Tuple[int]]):
        # Reshape all the data to an array of 3d coordinates
        V_a = V.reshape(1, -1, 3)
        P_a = P.reshape(-1, 1, 3)
        if V_axis is not None:
            fH_new_shape = np.append(np.delete(fH.shape, P_axis + V_axis),
                                     [P_a.shape[0], V_a.shape[1]])
        else:
            fH_new_shape = np.append(np.delete(fH.shape, P_axis),
                                     [P_a.shape[0], 1])
        # Reshape fH to match the P
        fH_a = fH
        fH_a = np.moveaxis(fH_a, P_axis, np.arange(-len(P_axis), 0))

        # Reshape fH to match the V
        if V_axis is not None:
            fH_a = np.moveaxis(fH_a, V_axis, np.arange(-len(V_axis), 0))

        fH_a = fH_a.reshape(fH_new_shape)
        fH_a = np.moveaxis(fH_a, 0, -3)    # move the axis of frequencies
        return V_a, P_a, fH_a


class RSD_parallel_propagator(Propagator):
    """
    Propagation using Rayleigh-Sommerfeld Diffraction for parallel planes
    """

    def __init__(self, P: np.ndarray, V: np.ndarray, w_a: np.ndarray):
        super().__init__()
        # Creates the kernel. It check parallel planes
        self.K_rsd = RSD_kernel(V, P, w_a)
        assert P.ndim == 3, "P has to contain the 3d coordinates in a plane"
        assert V.ndim == 3 or V.ndim == 4, \
            "V has to contain the 3d coordinates as volume or plane"
        self.__P = P
        self.__V = V
        if self.__V.ndim == 4:
            self.__V_z_idx = {}
            for hdz, V_z in enumerate(self.__V):
                self.__V_z_idx[str(V_z[0, 0])] = hdz
        self.__w_a = w_a
        self.__fH = None
        self.lock = Lock()

    def set_fH(self, fH, P_axis, V_axis):
        """
        Set the new impulse response value to performance the propagation
        """
        self.__fH = fH
        ffH = np.fft.fft2(fH, s=self.K_rsd.kernel_shape(), axes=P_axis)
        # Reorder the data to do easier the operations
        ffH = np.moveaxis(ffH, P_axis, [-2, -1])
        # Reorder the frecuencies
        self.ffH = np.moveaxis(ffH, 0, -3)
        if V_axis is not None:
            ffH = np.moveaxis(ffH, V_axis, np.arange(-len(V_axis), 0))

    def propagate(self, fH: np.ndarray, P: np.ndarray,
                  V: np.ndarray,  wl_v: np.ndarray,
                  P_axis: Union[int, Tuple[int]] = None,
                  V_axis: Union[int, Tuple[int]] = None) -> np.ndarray:
        """
        Override of propagate in Propagate. Optimized for parallel planes
        """
        # Check the data and types
        assert P is self.__P, "P is not the same used for initialization"
        assert wl_v is self.__w_a, \
            "wl_v is not the same used for initialization"
        if V.ndim == 3 and self.__V.ndim == 4:
            assert V.base is self.__V, \
                "V is not the part of the one used for initialization"
            idx_plane = [self.__V_z_idx[str(V[0, 0])]]
        elif V.ndim == 3 and self.__V.ndim == 3:
            assert V is self.__V, \
                "V is not the same used for initialization"
            idx_plane = [0]
        elif V.ndim == 4:
            assert V is self.__V, \
                "V is not the same used for initialization"
            idx_plane = range(V.shape[0])
        else:
            raise "propagate does not support this data"

        # Check the axis
        if P_axis is None:
            P_l_axis = [-2, -1]
        else:
            P_l_axis = P_axis

        with self.lock:
            if self.__fH is None or self.__fH is not fH:
                self.set_fH(fH, P_l_axis, V_axis)

        # Space to store results
        fI_pre_shape = tuple(np.array(self.ffH.shape)[:-3])
        if V.ndim == 3:
            fI_shape = fI_pre_shape + (len(wl_v), 1, V.shape[0], V.shape[1])
        else:   # V.ndim == 4
            fI_shape = fI_pre_shape + (len(wl_v), V.shape[0],
                                       V.shape[1], V.shape[2])
        fI = np.zeros(fI_shape, dtype=np.complex128)

        # Propagate with the RSD kernel
        for idx_fI, plane_id in enumerate(idx_plane):
            K = self.K_rsd.get_f_RSD_kernel_i(plane_id)
            fI_padded = np.fft.ifft2(K * self.ffH)
            fI[..., idx_fI, :, :] = fI_padded[...,
                                              -fI_shape[-2]:, -fI_shape[-1]:]

        # Squeeze unnedded axes
        if V.ndim == 3:
            fI = np.squeeze(fI, axis=-3)

        return RSD_parallel_propagator.reshape_result(fI, fH, V, -V.ndim,
                                                      P_l_axis, V_axis)



"""
tal.render
======

Functions for simulating NLOS captured data.

This is a private module, it is recommended to use the command line interface instead of calling these functions directly.

See tal render -h for more information.
"""


def create_nlos_scene(folder_name, args):
    """
    It is recommended to use the command line interface instead of calling this function directly.

    See tal render -h for more information. You probably want to do tal render new <folder_name>
    """
    from tal.render import create
    create.create_nlos_scene(folder_name, args)


def render_nlos_scene(config_path, args):
    """
    It is recommended to use the command line interface instead of calling this function directly.

    See tal render -h for more information. You probably want to do tal render <scene>, or tal render new <folder_name>
    """
    from tal.render import render
    render.render_nlos_scene(config_path, args)
